#include <JGL/JGL.h>

#if __linux__
    #include <freetype2/ft2build.h>
    #include FT_FREETYPE_H
#endif
#if _WIN32
    #include <ft2build.h>
    #include FT_FREETYPE_H
#endif

namespace JGL {
    FT_Library ft;

    struct Font {
        int index = 0;
        FT_Face face;
    };
    std::vector<Font> faces;

    int LoadFont(const std::string &font_path) {
        if (ft == nullptr)
            return -1;

        Font font;
        if (FT_New_Face(ft, font_path.c_str(), 0, &font.face)) {
            std::cout << "Error::FREETYPE: Failed to load font!" << std::endl;
            return -1;
        }
        unsigned int newIndex = 0;
        for (const auto& f : faces)
            if (f.index >= newIndex)
                newIndex = f.index + 1;

        font.index = newIndex;
        faces.push_back(font);
        std::cout << "Loaded font from " << font_path << " with index " << newIndex << std::endl;
        return newIndex;
    }

    bool InitTextEngine() {
        if (FT_Init_FreeType(&ft))
            return true;
        return false;
    }

    void UnloadFont(int font_index) {
        for (int i = 0; i < faces.size(); i++)
            if (faces[i].index == font_index)
                FT_Done_Face(faces[i].face),
                faces.erase(faces.begin() + i);
    }
    FontCache fontCache;

    void PurgeFontCache() {
        fontCache.purgeCache();
    }

    //TODO multi-texturing for 8x speedup. I tried. We'll come back to it later.
    void J2D::DrawString(const Color4& color, const std::string& text, float x, float y, float scale, u32 size, unsigned int font_index) {
        glUseProgram(0); // Fixed-function pipeline.

        Font font{};
        CachedFont* cachedFont = fontCache.getFont(size, font_index);

        //If the font doesn't exist in the cache yet.
        if (!cachedFont) {
            fontCache.newFont(size, font_index);
            cachedFont = fontCache.getFont(size, font_index);
        }

        //Set up the regular font.
        for (const auto& f : faces)
            if (f.index == font_index)
                font = f;
        if (font.face == nullptr)
            return;

        FT_Set_Pixel_Sizes(font.face, 0, size);

        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);

        std::vector<GLuint> textures(text.length());

        //For each character
        for (int i = 0; i < text.length(); i++) {
            float x2, y2, w, h;
            //If the font is in the cache already.
            if (cachedFont->getGlyph(text.c_str()[i])) {
                CachedGlyph* glyph = cachedFont->getGlyph(text.c_str()[i]);

                glBindTexture(GL_TEXTURE_2D, *glyph->getTexture());
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);

                x2 = x + glyph->x2offset * scale;
                y2 = y - glyph->y2offset * scale; // Adjust y-coordinate
                w = glyph->w * scale;
                h = glyph->h * scale;
                x += glyph->advanceX * scale;
                y += glyph->advanceY * scale;

            } else {
                if (FT_Load_Char(font.face, text.c_str()[i], FT_LOAD_RENDER))
                    continue;

                FT_GlyphSlot g = font.face->glyph;
                glGenTextures(1, &textures.at(i));
                glBindTexture(GL_TEXTURE_2D, textures[i]);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
                glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, g->bitmap.width, g->bitmap.rows, 0, GL_ALPHA, GL_UNSIGNED_BYTE, g->bitmap.buffer);

                x2 = x + g->bitmap_left * scale;
                y2 = -y - g->bitmap_top * scale; // Adjust y-coordinate
                w = g->bitmap.width * scale;
                h = g->bitmap.rows * scale;
                x += (g->advance.x >> 6) * scale;
                y += (g->advance.y >> 6) * scale;
                cachedFont->appendGlyph(new CachedGlyph(textures.at(i), text.c_str()[i], g->bitmap_left, g->bitmap_top, g->bitmap.width, g->bitmap.rows, (g->advance.x >> 6), (g->advance.y >> 6)));
            }

            GLfloat vertices[12] = {x2, y2, x2, y2 + h, x2 + w, y2 + h,x2, y2, x2 + w, y2 + h, x2 + w, y2};
            GLfloat textureCoordinates[12] = {0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0};

            glTexCoordPointer(2, GL_FLOAT, sizeof(GL_FLOAT) * 2, &textureCoordinates);
            glVertexPointer(2, GL_FLOAT, sizeof(GL_FLOAT) * 2, &vertices);
            glDrawArrays(GL_TRIANGLES, 0, 6);
        }
        glBindTexture(GL_TEXTURE_2D, 0); // Unbind texture
    }

    void J2D::DrawString(const Color3& color, const std::string& text, float x, float y, float scale, u32 size, unsigned int font_index) {
        J2D::DrawString(Color4::FromColor3(color, 255), text, x, y, scale, size, font_index);
    }

    void J3D::DrawString(const Color3& color, const std::string& text, const Vector3& pos, const Vector3& angle, float scale, u32 size, unsigned int font_index) {
        //TODO figure out what the scale should actually be mathematically.
        scale = scale * 0.002f;
        scale = -scale;

        float x = pos.x;
        float y = pos.y;
        float z = pos.z;
        std::vector<GLuint> textures(text.length());;
        glUseProgram(0); // Fixed-function pipeline.
        glColor4f(color.r, color.g, color.b, 1.0f);

        Font font;
        for (auto& f : faces)
            if (f.index == font_index)
                font = f;
        if (font.face == NULL) {
            std::cout << "null font" << std::endl;
            return;
        }

        FT_Set_Pixel_Sizes(font.face, 0, size);

        glPushMatrix();
        glTranslatef(x, y, z);
        glRotatef(angle.x, 1.0f, 0.0f, 0.0f);
        glRotatef(angle.y, 0.0f, 1.0f, 0.0f);
        glRotatef(angle.z, 0.0f, 0.0f, 1.0f);
        x = 0;
        y = 0;
        z = 0;

        for (int i = 0; i < text.length(); i++)
        {
            if (FT_Load_Char(font.face, text.c_str()[i], FT_LOAD_RENDER))
                continue;

            FT_GlyphSlot g = font.face->glyph;
            glGenTextures(1, &textures.at(i));
            glBindTexture(GL_TEXTURE_2D, textures[i]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, g->bitmap.width, g->bitmap.rows, 0, GL_ALPHA, GL_UNSIGNED_BYTE, g->bitmap.buffer);

            float x2 = x + g->bitmap_left * scale;
            float y2 = y - g->bitmap_top * scale; // Adjust y-coordinate
            float z2 = z;
            float w = g->bitmap.width * scale;
            float h = g->bitmap.rows * scale;

            glBegin(GL_TRIANGLES);

            glTexCoord2f(0, 0);
            glVertex3f(x2, y2, z2);

            glTexCoord2f(0, 1);
            glVertex3f(x2, y2 + h, z2);

            glTexCoord2f(1, 1);
            glVertex3f(x2 + w, y2 + h, z2);

            glTexCoord2f(0, 0);
            glVertex3f(x2, y2, z2);

            glTexCoord2f(1, 1);
            glVertex3f(x2 + w, y2 + h, z2);

            glTexCoord2f(1, 0);
            glVertex3f(x2 + w, y2, z2);

            glEnd();

            x += (g->advance.x >> 6) * scale;
            y += (g->advance.y >> 6) * scale;
        }

        for (unsigned int& texture : textures)
            glDeleteTextures(1, &texture);

        glBindTexture(GL_TEXTURE_2D, 0); // Unbind texture

        glPopMatrix();
    }
}
